import asyncio
import copy
import json
import os
import shlex
import subprocess
import sys
from datetime import datetime
from json.decoder import JSONDecodeError
from typing import Any, Dict, List, Optional, Union, cast

import pymongo
import tornado.httpclient
from watchdog import observers
from watchdog.events import FileSystemEvent, FileSystemEventHandler

import kekmonitors.utils.tools
from kekmonitors.config import COMMANDS, ERRORS, Config, LogConfig
from kekmonitors.utils.server.msg import Cmd, Response, badResponse, okResponse
from kekmonitors.utils.server.server import Server
from kekmonitors.utils.tools import list_contains_find_item


def get_directory_from_file(src: str):
	return src[:src.rfind(os.path.sep)]


class MonitorManager(Server, FileSystemEventHandler):
	'''This can be used to manage monitors/scrapers with an external api.'''

	def __init__(self, config: Config = Config()):
		if not config.name:
			config.name = f"Executable.MonitorManager"

		self.config = config

		super().__init__(config, f"{self.config.socket_path}/MonitorManager")
		super(Server).__init__()
		logconfig = LogConfig(self.config)
		logconfig.name += ".General"
		self.general_logger = kekmonitors.utils.tools.get_logger(logconfig)
		self.cmd_to_callback[COMMANDS.MM_STOP_MONITOR_MANAGER] = self._stop_serving
		self.cmd_to_callback[COMMANDS.MM_ADD_MONITOR] = self.on_add_monitor
		self.cmd_to_callback[COMMANDS.MM_ADD_SCRAPER] = self.on_add_scraper
		self.cmd_to_callback[COMMANDS.MM_ADD_MONITOR_SCRAPER] = self.on_add_monitor_scraper
		self.cmd_to_callback[COMMANDS.MM_STOP_MONITOR] = self.on_stop_monitor
		self.cmd_to_callback[COMMANDS.MM_STOP_SCRAPER] = self.on_stop_scraper
		self.cmd_to_callback[COMMANDS.MM_STOP_MONITOR_SCRAPER] = self.on_stop_monitor_scraper
		self.cmd_to_callback[COMMANDS.MM_GET_MONITOR_STATUS] = self.on_get_monitor_status
		self.cmd_to_callback[COMMANDS.MM_GET_SCRAPER_STATUS] = self.on_get_scraper_status
		self.cmd_to_callback[COMMANDS.MM_GET_MONITOR_SCRAPER_STATUS] = self.on_get_status
		self.cmd_to_callback[COMMANDS.MM_GET_MONITOR_CONFIG] = self.on_get_monitor_config
		self.cmd_to_callback[COMMANDS.MM_SET_MONITOR_CONFIG] = self.on_set_monitor_config
		self.cmd_to_callback[COMMANDS.MM_GET_MONITOR_WHITELIST] = self.on_get_monitor_whitelist
		self.cmd_to_callback[COMMANDS.MM_SET_MONITOR_WHITELIST] = self.on_set_monitor_whitelist
		self.cmd_to_callback[COMMANDS.MM_GET_MONITOR_BLACKLIST] = self.on_get_monitor_blacklist
		self.cmd_to_callback[COMMANDS.MM_SET_MONITOR_BLACKLIST] = self.on_set_monitor_blacklist
		self.cmd_to_callback[COMMANDS.MM_GET_MONITOR_WEBHOOKS] = self.on_get_monitor_webhooks
		self.cmd_to_callback[COMMANDS.MM_SET_MONITOR_WEBHOOKS] = self.on_set_monitor_webhooks
		self.cmd_to_callback[COMMANDS.MM_GET_SCRAPER_CONFIG] = self.on_get_scraper_config
		self.cmd_to_callback[COMMANDS.MM_SET_SCRAPER_CONFIG] = self.on_set_scraper_config
		self.cmd_to_callback[COMMANDS.MM_GET_SCRAPER_WHITELIST] = self.on_get_scraper_whitelist
		self.cmd_to_callback[COMMANDS.MM_SET_SCRAPER_WHITELIST] = self.on_set_scraper_whitelist
		self.cmd_to_callback[COMMANDS.MM_GET_SCRAPER_BLACKLIST] = self.on_get_scraper_blacklist
		self.cmd_to_callback[COMMANDS.MM_SET_SCRAPER_BLACKLIST] = self.on_set_scraper_blacklist
		self.cmd_to_callback[COMMANDS.MM_GET_SCRAPER_WEBHOOKS] = self.on_get_scraper_webhooks
		self.cmd_to_callback[COMMANDS.MM_SET_SCRAPER_WEBHOOKS] = self.on_set_scraper_webhooks

		self.monitor_processes = {}  # type: Dict[str, Dict[str, Any]]
		self.scraper_processes = {}  # type: Dict[str, Dict[str, Any]]
		self.monitor_sockets = {}  # type: Dict[str, str]
		self.scraper_sockets = {}  # type: Dict[str, str]
		self.socket_lock = asyncio.Lock()

		# mandatory arguments, needed in the command
		self.db = pymongo.MongoClient(self.config.db_path)[
                    self.config.db_name]["register"]
		self.add_args = ["name"]
		self.stop_args = ["name"]
		self.getter_configs_args = ["name"]
		self.setter_configs_args = ["name", "payload"]
		self.shutdown_all_on_exit = True   # you might wanna change this

		self._loop_lock = asyncio.Lock()
		tornado.httpclient.AsyncHTTPClient.configure(
			"tornado.curl_httpclient.CurlAsyncHTTPClient")
		self.client = tornado.httpclient.AsyncHTTPClient()

		# watches the config folder for any change. calls on_modified when any monitored file is modified
		self.config_watcher = observers.Observer()
		self.config_watcher.schedule(self, self.config.config_path, True)
		self.config_watcher.schedule(self, self.config.socket_path, True)
		self.has_to_quit = False

	def start(self):
		'''Start the Monitor Manager.'''
		self.config_watcher.start()
		self.check_status_task = self._asyncio_loop.create_task(self.check_status())
		self._asyncio_loop.run_forever()

	def on_modified(self, event: FileSystemEvent):
		# called when any of the monitored files is modified.
		# we are only interested int the configs for now.
		filename = event.key[1]  # type: str
		# if a config file is updated:
		splits = filename.split(os.path.sep)
		if splits[-1].endswith(".json"):
			if list_contains_find_item(splits, "config"):
				asyncio.run_coroutine_threadsafe(
					self.update_configs(filename), self._asyncio_loop)

	def on_created(self, event: FileSystemEvent):
		filename = event.key[1]  # type: str
		splits = filename.split(os.path.sep)
		if get_directory_from_file(filename) == self.config.socket_path:
			asyncio.run_coroutine_threadsafe(self.on_add_sockets(), self._asyncio_loop)

	def on_deleted(self, event):
		filename = event.key[1]  # type: str
		filename.split(os.path.sep)
		if get_directory_from_file(filename) == self.config.socket_path:
			asyncio.run_coroutine_threadsafe(
				self.on_delete_sockets(), self._asyncio_loop)

	async def on_add_sockets(self):
		async with self.socket_lock:
			new_monitor_sockets = {}  # type: Dict[str, str]
			new_scraper_sockets = {}  # type: Dict[str, str]
			for filename in os.listdir(self.config.socket_path):
				splits = filename.split(".")
				if splits[0] == "Monitor" and splits[1] not in self.monitor_sockets:
					new_monitor_sockets[splits[1]] = os.path.sep.join(
						[self.config.socket_path, filename])
				elif splits[0] == "Scraper" and splits[1] not in self.scraper_sockets:
					new_scraper_sockets[splits[1]] = os.path.sep.join(
						[self.config.socket_path, filename])

			alive_monitor_sockets, alive_scraper_sockets = await asyncio.gather(self.get_alive_sockets(new_monitor_sockets.values()), self.get_alive_sockets(new_scraper_sockets.values()))

			for class_name in new_monitor_sockets:
				if new_monitor_sockets[class_name] in alive_monitor_sockets:
					self.monitor_sockets[class_name] = new_monitor_sockets[class_name]

			for class_name in new_scraper_sockets:
				if new_scraper_sockets[class_name] in alive_scraper_sockets:
					self.scraper_sockets[class_name] = new_scraper_sockets[class_name]

	async def on_delete_sockets(self):
		async with self.socket_lock:
			monitor_sockets = []
			scraper_sockets = []
			for f in list(os.listdir(self.config.socket_path)):
				if f.startswith("Monitor."):
					monitor_sockets.append(f)
				elif f.startswith("Scraper."):
					scraper_sockets.append(f)

			new_monitor_sockets = copy.deepcopy(self.monitor_sockets)
			new_scraper_sockets = copy.deepcopy(self.scraper_sockets)

			for class_name in self.monitor_sockets:
				if "Monitor." + class_name not in monitor_sockets:
					new_monitor_sockets.pop(class_name)
			for class_name in self.scraper_sockets:
				if "Scraper." + class_name not in scraper_sockets:
					new_scraper_sockets.pop(class_name)

			self.monitor_sockets = new_monitor_sockets
			self.scraper_sockets = new_scraper_sockets

	async def get_alive_sockets(self, sockets: List[str]) -> List[str]:
		tasks = []
		for socket in sockets:
			cmd = Cmd()
			cmd.cmd = COMMANDS.PING
			tasks.append(self.make_request(socket, cmd))

		responses = await asyncio.gather(*tasks)  # type: List[Response]
		alive = []
		for response, socket in zip(responses, sockets):
			if not response.error.value:
				alive.append(socket)

		return alive

	async def update_configs(self, filename: str):
		'''Reads the config file and updates the interested monitors/scrapers'''
		self.general_logger.debug(f"File {filename} has changed!")
		if filename.endswith(".json"):
			try:
				with open(filename, "r") as f:
					j = json.load(f)
			except JSONDecodeError:
				self.general_logger.warning(
					f"File {filename} was changed but constains invalid json data")
				return

			splits = filename.split(os.path.sep)
			commands = []  # List[Cmd]
			cmd = None  # Enum
			sock_paths = []   # type: List[str]
			# if it's from the monitors folder:
			if list_contains_find_item(splits, "monitors"):
				# we are interested in configs, whitelist, blacklist, webhooks
				if splits[-1] == "whitelists.json":
					cmd = COMMANDS.SET_WHITELIST
				elif splits[-1] == "configs.json":
					cmd = COMMANDS.SET_CONFIG
				elif splits[-1] == "blacklists.json":
					cmd = COMMANDS.SET_BLACKLIST
				elif splits[-1] == "webhooks.json":
					cmd = COMMANDS.SET_WEBHOOKS
				else:
					cmd = None
				if cmd:
					# for every monitor socket
					for sockname in os.listdir(self.config.socket_path):
						if sockname.startswith("Monitor."):
							# add command and socket path to the list of todo
							name = sockname.split(".")[1]
							if (name in j):
								c = Cmd()
								c.cmd = cmd
								# send only the corresponding part to the monitor
								c.payload = j[name]
								commands.append(c)
								sock_paths.append(os.path.sep.join(
									[self.config.socket_path, sockname]))

			elif list_contains_find_item(splits, "scrapers"):
				# we are interested in configs, whitelist, blacklist
				if splits[-1] == "whitelists.json":
					cmd = COMMANDS.SET_WHITELIST
				elif splits[-1] == "configs.json":
					cmd = COMMANDS.SET_CONFIG
				elif splits[-1] == "blacklists.json":
					cmd = COMMANDS.SET_BLACKLIST
				else:
					cmd = None
				if cmd:
					# for every scraper socket
					for sockname in os.listdir(self.config.socket_path):
						if sockname.startswith("Scraper."):
							# add command and socket path to the list of todo
							name = sockname.split(".")[1]
							if (name in j):
								c = Cmd()
								c.cmd = cmd
								# send only the corresponding part to the scraper
								c.payload = j[name]
								commands.append(c)
								sock_paths.append(os.path.sep.join(
									[self.config.socket_path, sockname]))
			else:
				self.general_logger.debug("File not useful.")
				return

			# prepare to make all the async requests
			tasks = []
			for sock_path, command in zip(sock_paths, commands):
				tasks.append(self.make_request(sock_path, command))

			# send the requests
			responses = await asyncio.gather(*tasks)   # List[Response]

			for response in responses:
				if response.error.value:
					self.general_logger.warning(f"Failed to update config: {response.error}")

	async def on_server_stop(self):
		async with self._loop_lock:
			# stop the config watcher
			self.config_watcher.stop()
			self.config_watcher.join()

			if self.shutdown_all_on_exit:
				# get all the existing sockets
				sockets = []  # type: List[str]
				tasks = []
				for sockname in os.listdir(self.config.socket_path):
					if sockname.startswith("Scraper.") or sockname.startswith("Monitor."):
						cmd = Cmd()
						cmd.cmd = COMMANDS.STOP
						sockets.append(sockname)
						self.general_logger.info(f"Stopping {sockname}...")
						tasks.append(self.make_request(
							f"{self.config.socket_path}{os.path.sep}{sockname}", cmd))

				# send request to stop
				responses = await asyncio.gather(*tasks)   # type: List[Response]

				for sockname, r in zip(sockets, responses):
					# if an error happened...
					if r.error.value:
						# if the socket was not used remove it
						if r.error == ERRORS.SOCKET_COULDNT_CONNECT:
							os.remove(os.path.sep.join([self.config.socket_path, sockname]))
							self.general_logger.info(
								f"{self.config.socket_path}{os.path.sep}{sockname} was removed because unavailable")
						# else something else happened, dont do anything
						else:
							self.general_logger.warning(
								f"Error occurred while attempting to stop {sockname}: {r.error}")
					# ok
					else:
						self.general_logger.warning(f"{sockname} was successfully stopped")

		self._asyncio_loop.stop()
		self.general_logger.info("Shutting down...")
		return okResponse()

	async def check_status(self):
		'''Main MonitorManager loop. Every second it checks its monitored processes and looks if they are still alive, possibly reporting any exit code'''
		while True:
			async with self._loop_lock:
				new_monitor_processes = {}
				for class_name in self.monitor_processes:
					monitor = self.monitor_processes[class_name]["process"]
					if monitor.poll() is not None:
						log = f"Monitor {class_name} has stopped with code: {monitor.returncode}"
						if monitor.returncode:
							self.general_logger.warning(log)
							if self.config.crash_webhook:
								data = {"content": log}
								await self.client.fetch(self.config.crash_webhook, method="POST", body=json.dumps(data), headers={"content-type": "application/json"}, raise_error=False)
						else:
							self.general_logger.info(log)
					else:
						new_monitor_processes[class_name] = self.monitor_processes[class_name]
				self.monitor_processes = new_monitor_processes

				new_scraper_processes = {}
				for class_name in self.scraper_processes:
					scraper = self.scraper_processes[class_name]["process"]
					if scraper.poll() is not None:
						log = f"Scraper {class_name} has stopped with code: {scraper.returncode}"
						if scraper.returncode:
							self.general_logger.warning(log)
							if self.config.crash_webhook:
								data = {"content": log}
								await self.client.fetch(self.config.crash_webhook, method="POST", body=json.dumps(data), headers={"content-type": "application/json"}, raise_error=False)
						else:
							self.general_logger.info(log)
					else:
						new_scraper_processes[class_name] = self.scraper_processes[class_name]
				self.scraper_processes = new_scraper_processes
			await asyncio.sleep(1)

	async def on_add_monitor(self, cmd: Cmd) -> Response:
		r = badResponse()
		success, missing = cmd.has_valid_args(self.add_args)
		if success:
			payload = cast(Dict[str, Any], cmd.payload)
			db_monitor = self.db["monitors"].find_one({"name": payload["name"]})
			if db_monitor:
				success, reason = await self.add_monitor(
					db_monitor["path"], payload["name"], payload.get("delay", None))
				if success:
					r = okResponse()
				else:
					r.error = ERRORS.MM_COULDNT_ADD_MONITOR
					r.info = reason
			else:
				r.error = ERRORS.MONITOR_NOT_REGISTERED
				r.info = f"Tried to add monitor {payload['name']} but it was not found in the db. Did you start it at least once manually?"
		else:
			r.error = ERRORS.MISSING_PAYLOAD_ARGS
			r.info = f"Missing arguments: {missing}"
		return r

	async def on_add_scraper(self, cmd: Cmd) -> Response:
		r = badResponse()
		success, missing = cmd.has_valid_args(self.add_args)
		if success:
			payload = cast(Dict[str, Any], cmd.payload)
			db_scraper = self.db["scrapers"].find_one({"name": payload["name"]})
			if db_scraper:
				success, reason = await self.add_scraper(
					db_scraper["path"], payload["name"], payload.get("delay", None))
				if success:
					r = okResponse()
				else:
					r.error = ERRORS.MM_COULDNT_ADD_SCRAPER
					r.info = reason
			else:
				r.error = ERRORS.SCRAPER_NOT_REGISTERED
				r.info = f"Tried to add scraper {payload['name']} but it was not found in the db. Did you start it at least once manually?"
		else:
			r.error = ERRORS.MISSING_PAYLOAD_ARGS
			r.info = f"Missing arguments: {missing}"

		return r

	async def on_add_monitor_scraper(self, cmd: Cmd) -> Response:
		r = Response()
		r1, r2 = await asyncio.gather(self.on_add_monitor(cmd), self.on_add_scraper(cmd))
		r.error = ERRORS.OK if not r1.error.value and not r2.error.value else ERRORS.MM_COULDNT_ADD_MONITOR_SCRAPER
		r.info = f"Monitor: {r1.error.name}, Scraper: {r2.error.name}"
		if r.error.value and r.error:
			self.general_logger.warning(f"Couldn't add monitor and scraper")
			kekmonitors.utils.tools.dump_error(self.general_logger, r)

		return r

	async def on_stop_monitor(self, cmd: Cmd) -> Response:
		r = badResponse()
		success, missing = cmd.has_valid_args(self.stop_args)
		if success:
			payload = cast(Dict[str, Any], cmd.payload)
			socket = f"{self.config.socket_path}/Monitor.{payload['name']}"
			command = Cmd()
			command.cmd = COMMANDS.STOP
			self.general_logger.debug(f"Sending STOP to {socket}...")
			r = await self.make_request(socket, command)
			self.general_logger.debug(f"Sent STOP to {socket}")
		else:
			r.error = ERRORS.MISSING_PAYLOAD_ARGS
			r.info = f"Missing arguments: {missing}"
		return r

	async def on_stop_scraper(self, cmd: Cmd) -> Response:
		r = badResponse()
		success, missing = cmd.has_valid_args(self.stop_args)
		if success:
			payload = cast(Dict[str, Any], cmd.payload)
			socket = f"{self.config.socket_path}/Scraper.{payload['name']}"
			command = Cmd()
			command.cmd = COMMANDS.STOP
			self.general_logger.debug(f"Sending STOP to {socket}...")
			r = await self.make_request(socket, command)
			self.general_logger.debug(f"Sent STOP to {socket}")
		else:
			r.error = ERRORS.MISSING_PAYLOAD_ARGS
			r.info = f"Missing arguments: {missing}"
		return r

	async def on_stop_monitor_scraper(self, cmd: Cmd) -> Response:
		r = badResponse()
		success, missing = cmd.has_valid_args(self.stop_args)
		if success:
			r1, r2 = await asyncio.gather(self.on_stop_monitor(cmd), self.on_stop_scraper(cmd))
			r.error = ERRORS.OK if not r1.error.value and not r2.error.value else ERRORS.MM_COULDNT_STOP_MONITOR_SCRAPER
			r.info = f"Monitor: {r1.error.name}, Scraper: {r2.error.name}"
			if r.error.value and r.error:
				self.general_logger.warning(f"Couldn't stop monitor and scraper")
				kekmonitors.utils.tools.dump_error(self.general_logger, r)
		else:
			r.error = ERRORS.MISSING_PAYLOAD_ARGS
			r.info = f"Missing arguments: {missing}"
		return r

	async def on_get_monitor_status(self, cmd: Cmd) -> Response:
		process_status = {}
		for class_name in self.monitor_processes:
			start = self.monitor_processes[class_name]["start"].strftime(
				"%m/%d/%Y, %H:%M:%S")
			process_status[class_name] = {"Started at": start}
		sockets_status = {}
		# for class_name in self.monitor_sockets:
		#sockets_status[class_name] = {class_name: self.monitor_sockets[class_name]}
		sockets_status = self.monitor_sockets
		response = okResponse()
		response.payload = {
			"monitored_processes": process_status, "available_sockets": sockets_status}
		return response

	async def on_get_scraper_status(self, cmd: Cmd) -> Response:
		process_status = {}
		for class_name in self.scraper_processes:
			start = self.scraper_processes[class_name]["start"].strftime(
				"%m/%d/%Y, %H:%M:%S")
			process_status[class_name] = {"Started at": start}
		sockets_status = {}
		# for class_name in self.scraper_sockets:
		#sockets_status[class_name] = {class_name: self.scraper_sockets[class_name]}
		sockets_status = self.scraper_sockets
		response = okResponse()
		response.payload = {
			"monitored_processes": process_status, "available_sockets": sockets_status}
		return response

	async def on_get_status(self, cmd: Cmd) -> Response:
		ms = await self.on_get_monitor_status(cmd)
		ss = await self.on_get_scraper_status(cmd)
		response = okResponse()
		msp = cast(Dict[str, Any], ms.payload)  # type: Dict[str, Any]
		ssp = cast(Dict[str, Any], ss.payload)  # type: Dict[str, Any]
		response.payload = {
			"monitors": msp, "scrapers": ssp}
		return response

	async def add_monitor(self, filename: str, class_name: str, delay: Optional[Union[int, float]]):
		if class_name in self.monitor_processes:
			self.general_logger.debug(
				f"Tried to add an already existing monitor ({class_name} ({filename}))")
			return False, "Monitor already started."

		cmd = f"nohup {sys.executable} {filename} --no-output"
		if delay:
			cmd += f" --delay {str(delay)}"

		self.general_logger.debug(f"Starting {class_name} ({filename})...")
		monitor = subprocess.Popen(shlex.split(
			cmd), stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

		await asyncio.sleep(2)  # wait to check if process is still alive

		if monitor.poll() is not None:
			success = False
			msg = f"Failed to start monitor {class_name} ({filename})"
			self.general_logger.warning(
				f"Tried to start {class_name} ({filename}) but failed")
		else:
			self.general_logger.info(
				f"Added monitor {class_name} with pid {monitor.pid}")
			self.monitor_processes[class_name] = {
				"process": monitor,
				"start": datetime.now()
			}
			success = True
			msg = ""
		return success, msg

	async def add_scraper(self, filename: str, class_name: str, delay: Optional[Union[int, float]]):
		if class_name in self.scraper_processes:
			self.general_logger.debug(
				f"Tried to add an already existing scraper ({class_name} ({filename}))")
			return False, "Scraper already started."

		cmd = f"nohup {sys.executable} {filename} --no-output"
		if delay:
			cmd += f" --delay {str(delay)}"

		self.general_logger.debug(f"Starting {class_name} ({filename})...")
		scraper = subprocess.Popen(shlex.split(
			cmd), stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

		await asyncio.sleep(2)  # wait to check if process is still alive

		if scraper.poll() is not None:
			success = False
			msg = f"Failed to start scraper {class_name} ({filename})"
			self.general_logger.warning(
				f"Tried to start {class_name} ({filename}) but failed")
		else:
			self.general_logger.info(
				f"Added scraper {class_name} with pid {scraper.pid}")
			self.scraper_processes[class_name] = {
				"process": scraper,
				"start": datetime.now()
			}
			success = True
			msg = ""
		return success, msg

	async def on_get_monitor_config(self, cmd: Cmd) -> Response:
		return await self.getter_config(cmd, COMMANDS.GET_CONFIG, True)

	async def on_set_monitor_config(self, cmd: Cmd) -> Response:
		return await self.setter_config(cmd, COMMANDS.SET_CONFIG, True)

	async def on_get_monitor_whitelist(self, cmd: Cmd) -> Response:
		return await self.getter_config(cmd, COMMANDS.GET_WHITELIST, True)

	async def on_set_monitor_whitelist(self, cmd: Cmd) -> Response:
		return await self.setter_config(cmd, COMMANDS.SET_WHITELIST, True)

	async def on_get_monitor_blacklist(self, cmd: Cmd) -> Response:
		return await self.getter_config(cmd, COMMANDS.GET_BLACKLIST, True)

	async def on_set_monitor_blacklist(self, cmd: Cmd) -> Response:
		return await self.setter_config(cmd, COMMANDS.SET_BLACKLIST, True)

	async def on_get_monitor_webhooks(self, cmd: Cmd) -> Response:
		return await self.getter_config(cmd, COMMANDS.GET_WEBHOOKS, True)

	async def on_set_monitor_webhooks(self, cmd: Cmd) -> Response:
		return await self.setter_config(cmd, COMMANDS.SET_WEBHOOKS, True)

	async def on_get_scraper_config(self, cmd: Cmd) -> Response:
		return await self.getter_config(cmd, COMMANDS.GET_CONFIG, False)

	async def on_set_scraper_config(self, cmd: Cmd) -> Response:
		return await self.setter_config(cmd, COMMANDS.SET_CONFIG, False)

	async def on_get_scraper_whitelist(self, cmd: Cmd) -> Response:
		return await self.getter_config(cmd, COMMANDS.GET_WHITELIST, False)

	async def on_set_scraper_whitelist(self, cmd: Cmd) -> Response:
		return await self.setter_config(cmd, COMMANDS.SET_WHITELIST, False)

	async def on_get_scraper_blacklist(self, cmd: Cmd) -> Response:
		return await self.getter_config(cmd, COMMANDS.GET_BLACKLIST, False)

	async def on_set_scraper_blacklist(self, cmd: Cmd) -> Response:
		return await self.setter_config(cmd, COMMANDS.SET_BLACKLIST, False)

	async def on_get_scraper_webhooks(self, cmd: Cmd) -> Response:
		return await self.getter_config(cmd, COMMANDS.GET_WEBHOOKS, False)

	async def on_set_scraper_webhooks(self, cmd: Cmd) -> Response:
		return await self.setter_config(cmd, COMMANDS.SET_WEBHOOKS, False)

	async def getter_config(self, cmd: Cmd, command: COMMANDS, is_monitor: bool):
		success, missing = cmd.has_valid_args(self.getter_configs_args)
		if success:
			payload = cast(Dict[str, Any], cmd.payload)
			c = Cmd()
			c.cmd = command
			r = await self.make_request(f"{self.config.socket_path}/{'Monitor' if is_monitor else 'Scraper'}.{payload['class_name']}", c)
			return r
		else:
			r = badResponse()
			r.error = ERRORS.MISSING_PAYLOAD_ARGS
			r.info = f"{missing}"
			return r

	async def setter_config(self, cmd: Cmd, command: COMMANDS, is_monitor: bool):
		success, missing = cmd.has_valid_args(self.setter_configs_args)
		if success:
			payload = cast(Dict[str, Any], cmd.payload)
			c = Cmd()
			c.cmd = command
			c.payload = payload["payload"]
			r = await self.make_request(f"{self.config.socket_path}/{'Monitor' if is_monitor else 'Scraper'}.{payload['class_name']}", c)
			return r
		else:
			r = badResponse()
			r.error = ERRORS.MISSING_PAYLOAD_ARGS
			r.info = f"{missing}"
			return r


if __name__ == "__main__":
	MonitorManager().start()



import json
from typing import Any, Dict, List, Optional, Tuple

from kekmonitors.config import COMMANDS, ERRORS


class Message(object):
	'''
	Represents a message to be sent. It can be initialized with bytes taken from msg.get_bytes().\n
	It automatically sets the class (and sub-class) variables to the value set in provided msg.\n
	'''

	def __init__(self, msg: Optional[bytes] = None):
		# update member variables by cycling through the first level of the json dict
		if isinstance(msg, bytes):
			try:
				j = json.loads(msg)    # type: Dict[str, Any]
				for key in self.__dict__:
					self.__dict__[key] = j.get(key, None)
			except:
				pass
		else:
			pass

	def get_json(self) -> Optional[Any]:
		'''Return a json representation of the message'''
		return self.__dict__

	def get_bytes(self) -> bytes:
		'''Return a bytes representation of the message'''
		return json.dumps(self.get_json()).encode("utf-8")


class Cmd(Message):
	'''
	Represents a command to be sent. It can be initialized with bytes taken from cmd.get_bytes().\n
	payload can be used to return data.
	'''

	def __init__(self, msg: Optional[bytes] = None):
		self.cmd = None  # type: Optional[int]
		self.payload = None  # type: Optional[Any]
		super().__init__(msg)

	@property
	def cmd(self):
		if self.__cmd is not None:
			return COMMANDS(self.__cmd)
		else:
			return None

	@cmd.setter
	def cmd(self, cmd):
		if isinstance(cmd, COMMANDS):
			self.__cmd = cmd.value
		elif isinstance(cmd, int) or cmd is None:
			self.__cmd = cmd
		else:
			raise Exception(f"Tried to set cmd to unsupported value ({type(cmd)})")

	def has_valid_args(self, args: List[str]) -> Tuple[bool, Optional[List[str]]]:
		'''Checks if the payload in the cmd has all the required arguments.\n
		Returns (success, missing_arguments)'''
		if self.payload:
			missing = []
			for needed_arg in args:
				if needed_arg not in self.payload:
					missing.append(needed_arg)
			if missing:
				return False, missing
			else:
				return True, None
		return False, args


class Response(Message):
	'''
	Represents a response to a command. It can be initialized with bytes taken from a response.get_bytes().\n
	The success is represented by error == ERRORS.OK; info can be set to a human-readable string to provide more information about the error.\n
	payload can be used to return data.
	'''

	def __init__(self, msg: Optional[bytes] = None):
		self.error = None  # type: Optional[int]
		self.info = None  # type: Optional[str]
		self.payload = None  # type: Optional[Dict[str, Any]]
		super().__init__(msg)

	@property
	def error(self):
		if self.__error is not None:
			return ERRORS(self.__error)
		else:
			return None

	@error.setter
	def error(self, error):
		if isinstance(error, ERRORS):
			self.__error = error.value
		elif isinstance(error, int) or error is None:
			self.__error = error
		else:
			raise Exception(f"Tried to set error to unsupported value ({type(error)})")

	@property
	def info(self):
		return self.__info

	@info.setter
	def info(self, info):
		if isinstance(info, str) or info is None:
			self.__info = info
		else:
			raise Exception(f"Tried to set info to unsupported value ({type(info)})")


def badResponse():
	'''Quickly create an unsuccessful response with a default error.'''
	r = Response()
	r.error = ERRORS.OTHER_ERROR
	return r


def okResponse():
	'''Quickly create a successful response'''
	r = Response()
	r.error = ERRORS.OK
	return r



import asyncio
import copy
import json
import traceback
from typing import List, Optional

from kekmonitors.base_common import Common
from kekmonitors.config import COMMANDS, ERRORS, Config
from kekmonitors.utils import discord_embeds, shoe_stuff
from kekmonitors.utils.network_utils import NetworkUtils
from kekmonitors.utils.server.msg import Cmd, Response, badResponse, okResponse
from kekmonitors.utils.server.server import Server
from kekmonitors.utils.shoe_manager import ShoeManager
from kekmonitors.utils.shoe_stuff import Shoe
from kekmonitors.utils.tools import dump_error
from kekmonitors.utils.webhook_manager import WebhookManager


class BaseMonitor(Common, NetworkUtils):
	def __init__(self, config: Config = Config()):
		if not config.name:
			config.name = f"Monitor.{self.get_class_name()}"
		elif not config.name.startswith("Monitor."):
			raise Exception(
				f"You must start the monitor name with \"Monitor.\"! Currently: {config.name}")
		self.crash_webhook = config.crash_webhook

		super().__init__(config)
		super(Server, self).__init__(config.name)

		self._mark_as_monitor()

		self.cmd_to_callback[COMMANDS.PING] = self._on_ping
		self.cmd_to_callback[COMMANDS.STOP] = self._stop_serving
		self.cmd_to_callback[COMMANDS.SET_LINKS] = self.on_set_links
		self.cmd_to_callback[COMMANDS.ADD_LINKS] = self.on_add_links

		self.buffer_links = []  # type: List[str]
		self.new_links = []  # type: List[str]
		self.links = []  # type: List[str]
		self.shoes = []  # type: List[Shoe]

		self.shoe_manager = ShoeManager()
		self.webhook_manager = WebhookManager(config)

		# website-specific variables should be declared here
		self.init()

	async def on_set_links(self, msg: Cmd) -> Response:
		response = badResponse()
		p = msg.payload
		if p is not None:
			if isinstance(p, list):
				self.new_links = p
				response.error = ERRORS.OK
			else:
				self.client_logger.warning(
					"Received new set of links, but payload was not a list: ", p)
				response.error = ERRORS.BAD_PAYLOAD
				response.info = f"Invalid links (expected list, got {type(p)}"

		else:
			self.client_logger.warning(
				"Missing payload; msg: ", msg.get_json())
			response.error = ERRORS.MISSING_PAYLOAD
		return response

	async def on_add_links(self, msg: Cmd) -> Response:
		response = badResponse()
		p = msg.payload
		if p is not None:
			if isinstance(p, list):
				self.buffer_links = p
				response.error = ERRORS.OK
			else:
				self.general_logger.warning(
					"Received new added links, but payload was not a list: ", p)
				response.error = ERRORS.BAD_PAYLOAD
				response.info = f"Invalid links (expected list, got {type(p)}"

		else:
			self.general_logger.warning(
				"Missing payload; msg: ", msg.get_json())
			response.error = ERRORS.MISSING_PAYLOAD
		return response

	async def on_server_stop(self) -> Response:
		self.general_logger.debug("Waiting for loop to complete...")
		async with self._loop_lock:
			pass
		self.general_logger.debug("Loop is completed, starting shutdown...")
		await self.on_async_shutdown()
		self._asyncio_loop.stop()
		self.general_logger.debug("Shutting down webhook manager...")
		self.webhook_manager.quit()
		self.on_shutdown()
		return okResponse()

	async def _on_ping(self, cmd: Cmd) -> Response:
		return okResponse()

	async def _get_links(self):
		socket_path = f"{self.config.socket_path}/Scraper.{self.class_name}"
		self.client_logger.debug("Getting links...")

		cmd = Cmd()
		cmd.cmd = COMMANDS.GET_LINKS
		response = await self.make_request(socket_path, cmd)
		if not response.error.value:
			if response.payload:
				self.links = response.payload
			else:
				self.client_logger.warning("Tried to get links but payload was invalid.")
				dump_error(self.client_logger, response)
		else:
			self.client_logger.warning(f"Failed to get links")
			dump_error(self.client_logger, response)

	async def main(self):
		'''Main loop. Updates configs, runs user-defined loop and performs links/shoes updates for the user'''
		await self.async_init()
		# Try to get a set of links as soon as the monitor starts.
		await self._get_links()
		while True:
			async with self._loop_lock:
				self.update_local_config()
				if self.new_links:
					self.general_logger.info(f"Received new set of links: {self.new_links}")
					self.links = self.new_links
					self.new_links = []
				if self.buffer_links:
					self.general_logger.info(f"Adding new set of links: {self.buffer_links}")
					self.links += self.buffer_links
					self.buffer_links = []
				try:
					self.shoes = []
					await self.loop()
					self.shoe_check()
				except:
					self.general_logger.exception("")
					if self.crash_webhook:
						data = json.dumps(
							{"content": f"{self.class_name} has crashed:\n{traceback.format_exc()}\nRestarting in {self.delay} secs."[:2000]})
						await self.client.fetch(self.crash_webhook, method="POST", body=data, headers={"content-type": "application/json"}, raise_error=False)
				self.general_logger.info(f"Loop ended. Waiting {self.delay} secs.")
			await asyncio.sleep(self.delay)

	async def loop(self):
		'''User-defined loop. Replace this with a function that will be run every `delay` seconds'''
		await asyncio.sleep(1)

	def shoe_check(self):
		'''This is run just after the user-defined loop. It compares the just found shoes with the ones in the db.\n
			Any changed shoe is sent to the webhooks.\n
			You probably want to override this function to set custom embeds.'''
		for shoe in self.shoes:
			returned = self.set_reason_and_update_shoe(shoe)
			if returned:
				embed = discord_embeds.get_default_embed(returned)
				self.webhook_manager.add_to_queue(embed, self.webhooks_json)

	def set_reason_and_update_shoe(self, shoe: Shoe) -> Optional[Shoe]:
		"""Check shoe against db. If present in db check if there are new sizes;\n
			if so, set reason to restock, update the db and return a shoe with only the restocked sizes;\n
			else update the shoe and return None. If not in the db return a copy of the shoe."""
		self.general_logger.debug(
			f"Checking {shoe.name} - {shoe.link} in db...")
		new_or_restocked = True
		db_shoe = self.shoe_manager.find_shoe({"link": shoe.link})
		return_shoe = copy.deepcopy(shoe)
		if db_shoe:
			new_or_restocked = False
			return_shoe.sizes = {}
			self.general_logger.debug("\tIt's present in db. Checking sizes.")
			self.general_logger.debug(f"\t\tdb_shoe sizes: {str(db_shoe.sizes)}")
			self.general_logger.debug(f"\t\tShoe sizes: {str(shoe.sizes)}")
			for shoe_size in shoe.sizes:
				try:
					db_available = db_shoe.sizes[shoe_size]["available"]
					if db_available != shoe.sizes[shoe_size]["available"]:
						if shoe.sizes[shoe_size]["available"] == True:
							new_or_restocked = True
							return_shoe.sizes[shoe_size] = shoe.sizes[shoe_size]
							self.general_logger.info(
								f"\t{shoe.link}: {str(shoe_size)} is now available")
				except (NameError, KeyError):
					# is shoe size available?
					if shoe.sizes[shoe_size]["available"]:
						new_or_restocked = True
						return_shoe.sizes[shoe_size] = shoe.sizes[shoe_size]
						self.general_logger.info(f"\t{shoe.link}: {str(shoe_size)} not in db.")

			if new_or_restocked:
				shoe.reason = shoe_stuff.RESTOCK
				return_shoe.reason = shoe_stuff.RESTOCK
				self.shoe_manager.update_shoe(shoe)
			else:
				self.general_logger.info(f"\t{shoe.link}: it has no restocked sizes.")
				self.shoe_manager.update_shoe(shoe)
				return None

		else:
			shoe.reason = shoe_stuff.NEW_RELEASE
			return_shoe.reason = shoe_stuff.NEW_RELEASE
			self.general_logger.info(f"\t{shoe.link}: not in db. Adding it.")
			self.shoe_manager.add_shoe(shoe)

		return return_shoe



from kekmonitors.config import COMMANDS
from kekmonitors.monitor_manager_cli import send
from kekmonitors.utils.server.msg import Cmd

if __name__ == "__main__":
	cmd = Cmd()
	cmd.cmd = COMMANDS.MM_STOP_MONITOR_MANAGER
	send(cmd)



import configparser
import enum
import logging
import os


@enum.unique
class COMMANDS(enum.Enum):
	PING = enum.auto()
	STOP = enum.auto()
	SET_LINKS = enum.auto()
	ADD_LINKS = enum.auto()
	GET_LINKS = enum.auto()
	SET_CONFIG = enum.auto()
	GET_CONFIG = enum.auto()
	SET_WHITELIST = enum.auto()
	GET_WHITELIST = enum.auto()
	SET_BLACKLIST = enum.auto()
	GET_BLACKLIST = enum.auto()
	SET_WEBHOOKS = enum.auto()
	GET_WEBHOOKS = enum.auto()

	MM_ADD_MONITOR = enum.auto()
	MM_ADD_SCRAPER = enum.auto()
	MM_ADD_MONITOR_SCRAPER = enum.auto()
	MM_STOP_MONITOR = enum.auto()
	MM_STOP_SCRAPER = enum.auto()
	MM_STOP_MONITOR_SCRAPER = enum.auto()
	MM_STOP_MONITOR_MANAGER = enum.auto()
	MM_GET_MONITOR_STATUS = enum.auto()
	MM_GET_SCRAPER_STATUS = enum.auto()
	MM_GET_MONITOR_SCRAPER_STATUS = enum.auto()
	MM_SET_MONITOR_CONFIG = enum.auto()
	MM_GET_MONITOR_CONFIG = enum.auto()
	MM_SET_MONITOR_WHITELIST = enum.auto()
	MM_GET_MONITOR_WHITELIST = enum.auto()
	MM_SET_MONITOR_BLACKLIST = enum.auto()
	MM_GET_MONITOR_BLACKLIST = enum.auto()
	MM_SET_MONITOR_WEBHOOKS = enum.auto()
	MM_GET_MONITOR_WEBHOOKS = enum.auto()
	MM_SET_SCRAPER_CONFIG = enum.auto()
	MM_GET_SCRAPER_CONFIG = enum.auto()
	MM_SET_SCRAPER_WHITELIST = enum.auto()
	MM_GET_SCRAPER_WHITELIST = enum.auto()
	MM_SET_SCRAPER_BLACKLIST = enum.auto()
	MM_GET_SCRAPER_BLACKLIST = enum.auto()
	MM_SET_SCRAPER_WEBHOOKS = enum.auto()
	MM_GET_SCRAPER_WEBHOOKS = enum.auto()


@enum.unique
class ERRORS(enum.Enum):
	OK = 0

	SOCKET_DOESNT_EXIST = enum.auto()
	SOCKET_COULDNT_CONNECT = enum.auto()
	SOCKET_TIMEOUT = enum.auto()

	MONITOR_DOESNT_EXIST = enum.auto()
	SCRAPER_DOESNT_EXIST = enum.auto()
	MONITOR_NOT_REGISTERED = enum.auto()
	SCRAPER_NOT_REGISTERED = enum.auto()

	UNRECOGNIZED_COMMAND = enum.auto()
	BAD_PAYLOAD = enum.auto()
	MISSING_PAYLOAD = enum.auto()
	MISSING_PAYLOAD_ARGS = enum.auto()

	MM_COULDNT_ADD_MONITOR = enum.auto()
	MM_COULDNT_ADD_SCRAPER = enum.auto()
	MM_COULDNT_ADD_MONITOR_SCRAPER = enum.auto()
	MM_COULDNT_STOP_MONITOR = enum.auto()
	MM_COULDNT_STOP_SCRAPER = enum.auto()
	MM_COULDNT_STOP_MONITOR_SCRAPER = enum.auto()

	OTHER_ERROR = enum.auto()
	UNKNOWN_ERROR = enum.auto()


# needed to avoid circular import
def get_file_if_exist_else_create(filename_path, content) -> str:
	filename_directory_path = filename_path[:filename_path.rfind(os.path.sep)]
	os.makedirs(filename_directory_path, exist_ok=True)
	if os.path.isfile(filename_path):
		with open(filename_path, "r") as rf:
			return rf.read()
	else:
		with open(filename_path, "w") as wf:
			wf.write(content)
		return content


class Config(object):
	def __init__(self):
		path = os.path.sep.join([os.environ['HOME'], ".kekmonitors", "config"])
		config_path = os.path.sep.join([path, "config.cfg"])
		default_config_str = f"\
[GlobalConfig]\n\
socket_path = {os.environ['HOME']}/.kekmonitors/sockets\n\
log_path = {os.environ['HOME']}/.kekmonitors/logs\n\
db_name = kekmonitors\n\
db_path = mongodb://localhost:27017/\n\
\n\
[DefaultBaseConfig]\n\
name = \n\
crash_webhook = \n\
provider = KekMonitors\n\
provider_icon = https://avatars0.githubusercontent.com/u/11823129?s=400&u=3e617374871087e64b5fde0df668260f2671b076&v=4\n\
timestamp_format = %d %b %Y, %H:%M:%S.%f\n\
embed_color = 255\n\
add_stream_handler = True\n\
loop_delay = 5\n"
		get_file_if_exist_else_create(config_path, default_config_str)
		parser = configparser.RawConfigParser()
		parser.read(config_path)
		parser.set("GlobalConfig", "config_path", path)

		self.config_path = parser.get("GlobalConfig", "config_path")
		self.socket_path = parser.get("GlobalConfig", "socket_path")
		self.log_path = parser.get("GlobalConfig", "log_path")
		self.db_name = parser.get("GlobalConfig", "db_name")
		self.db_path = parser.get("GlobalConfig", "db_path")

		self.name = parser.get("DefaultBaseConfig", "name")
		self.crash_webhook = parser.get("DefaultBaseConfig", "crash_webhook")
		self.provider = parser.get("DefaultBaseConfig", "provider")
		self.provider_icon = parser.get("DefaultBaseConfig", "provider_icon")
		self.timestamp_format = parser.get("DefaultBaseConfig", "timestamp_format")
		self.embed_color = int(parser.get("DefaultBaseConfig", "embed_color"))
		self.add_stream_handler = parser.get(
			"DefaultBaseConfig", "add_stream_handler")
		self.loop_delay = int(parser.get("DefaultBaseConfig", "loop_delay"))


class LogConfig(Config):
	def __init__(self, config: Config = None):
		if config is not None:
			for key in config.__dict__:
				self.__dict__[key] = config.__dict__[key]
		else:
			super().__init__()

		self.stream_level = logging.DEBUG
		self.file_level = logging.DEBUG



import asyncio
import copy
import json
import traceback
from typing import List

from kekmonitors.base_common import Common
from kekmonitors.config import COMMANDS, Config
from kekmonitors.utils.network_utils import NetworkUtils
from kekmonitors.utils.server.msg import Cmd, Message, Response, okResponse
from kekmonitors.utils.server.server import Server
from kekmonitors.utils.tools import dump_error


class BaseScraper(Common, NetworkUtils):
	def __init__(self, config: Config = Config()):
		if not config.name:
			config.name = f"Scraper.{self.get_class_name()}"
		elif not config.name.startswith("Scraper."):
			raise Exception(
				f"You must start the scraper name with \"Scraper.\"! Currently: {config.name}")
		self.crash_webhook = config.crash_webhook
		# init some internal variables (logger, links)

		super().__init__(config)
		super(Server, self).__init__(config.name)

		self._mark_as_scraper()

		self.cmd_to_callback[COMMANDS.PING] = self._on_ping
		self.cmd_to_callback[COMMANDS.STOP] = self._stop_serving
		self.cmd_to_callback[COMMANDS.GET_LINKS] = self.on_get_links
		self.links = []  # type: List[str]
		self._previous_links = []  # type: List[str]

		# website-specific variables should be declared here
		self.init()

	async def on_server_stop(self) -> Response:
		self.general_logger.debug("Waiting for loop to complete...")
		async with self._loop_lock:
			pass
		self.general_logger.debug("Loop is completed, starting shutdown...")
		await self.on_async_shutdown()
		self._asyncio_loop.stop()
		self.on_shutdown()
		return okResponse()

	async def on_get_links(self, cmd: Cmd) -> Message:
		response = okResponse()
		response.payload = self.links
		return response

	async def main(self):
		'''Main loop. Updates configs, runs user-defined loop and performs links/shoes updates for the user'''
		await self.async_init()
		while True:
			async with self._loop_lock:
				self.update_local_config()
				try:
					await self.loop()
					await self.update_links()
				except:
					self.general_logger.exception("")
					if self.crash_webhook:
						data = {"content": f"{self.class_name} has crashed:\n{traceback.format_exc()}\nRestarting in {self.delay} secs."}
						await self.client.fetch(self.crash_webhook, method="POST", body=json.dumps(data), headers={"content-type": "application/json"}, raise_error=False)
				self.general_logger.info(f"Loop ended, waiting {self.delay} secs")
			await asyncio.sleep(self.delay)

	async def loop(self):
		'''User-defined loop. Replace this with a function that will be run every `delay` seconds'''
		'''Modify self.links in order to force a link update on the corresponding monitor.'''
		await asyncio.sleep(1)

	async def update_links(self):
		'''This is called just after self.loop. Checks if any of the links have been modified and sends them to the corresponding monitor.'''
		if self.links != self._previous_links:
			await self._set_links()
			self._previous_links = copy.deepcopy(self.links)

	async def _on_ping(self, cmd: Cmd) -> Response:
		return okResponse()

	async def _set_links(self):
		'''Connect to the corresponding monitor, if available, and tell it to set the new links.'''
		socket_path = f"{self.config.socket_path}/Monitor.{self.class_name}"
		cmd = Cmd()
		cmd.cmd = COMMANDS.SET_LINKS
		cmd.payload = self.links
		response = await self.make_request(socket_path, cmd)
		if response.error.value:
			dump_error(self.client_logger, response)

	async def _add_links(self):
		'''Connect to the corresponding monitor, if available, and send it the new links.'''
		socket_path = f"{self.config.socket_path}/Monitor.{self.class_name}"
		cmd = Cmd()
		cmd.cmd = COMMANDS.ADD_LINKS
		cmd.payload = self.links
		response = await self.make_request(socket_path, cmd)
		if response.error.value:
			dump_error(self.client_logger, response)



import asyncio
import sys
from pprint import pprint

from kekmonitors.config import COMMANDS, Config
from kekmonitors.utils.server.msg import Cmd
from kekmonitors.utils.tools import make_request


def send(cmd: Cmd):
	print(f"Command: {cmd.cmd}")
	print(f"Payload: {cmd.payload}")
	print("Executing request...")
	response = asyncio.run(make_request(
		f"{Config().socket_path}/MonitorManager", cmd, True))

	print("E:", response.error.name)
	if response.info:
		print("Info:", response.info)
	if response.payload is not None:
		print("Payload:", end="")
		pprint(response.payload)


if __name__ == "__main__":
	args = sys.argv
	if len(args) < 2:
		print("Usage: python monitor_manager_cli.py <cmd> [payload]")
		print(f"To list available commands: python {args[0]} --list-cmd")
		exit(1)
	if args[1] == "--list-cmd":
		for key in COMMANDS.__dict__:
			try:
				COMMANDS[key]
				print(key)
			except:
				pass
		exit(0)
	cmd = COMMANDS.__dict__.get(args[1], None)
	if not cmd:
		print("Inserted cmd does not exist!")
		exit(2)
	string_cmd = args[1]
	payload = {}
	if len(args) > 2:
		if not len(args) % 2:
			for index, term in enumerate(args[2:]):
				if not index % 2:
					if not term.startswith("--"):
						print("You must start every payload key with \"--\"")
						exit(4)
					payload[term[2:]] = args[3 + index]
		else:
			print("Incorrect number of payload options!")
			exit(3)
	command = Cmd()
	command.cmd = cmd
	command.payload = payload
	send(command)



import asyncio
import enum
import os
from typing import Callable, Dict

from kekmonitors.config import ERRORS, Config, LogConfig
from kekmonitors.utils.server.msg import Cmd, Response, badResponse, okResponse
from kekmonitors.utils.tools import get_logger


class Server(object):
	def __init__(self, config: Config, server_path: str):
		logconfig = LogConfig(config)
		logconfig.name += ".Server"
		self.server_logger = get_logger(logconfig)
		self.server_path = server_path
		socket_directory = self.server_path[:self.server_path.rfind(os.path.sep)]
		try:
			os.makedirs(socket_directory, exist_ok=True)
		except PermissionError:
			self.server_logger.exception(
				"Failed to create the socket directory. Please provide a directory where the current user has write permissions!")
			exit(1)
		# init asyncio stuff
		self._asyncio_loop = asyncio.get_event_loop()
		self._server_task = self._asyncio_loop.create_task(self._init_server())

		self.cmd_to_callback = {}   # type: Dict[enum.Enum, Callable]

	async def _init_server(self):
		'''Initialise the underlying socket server, to allow communication between monitor/scraper.'''
		self.server = await asyncio.start_unix_server(self._handle_msg, self.server_path)

		addr = self.server.sockets[0].getsockname()
		self.server_logger.debug(f'Serving on {addr}')

	async def _stop_serving(self, msg: Cmd):
		self.server_logger.info("Closing server...")
		self.server.close()
		await self.server.wait_closed()
		os.remove(self.server_path)
		self.server_logger.info("Server successfully closed.")
		return await self.on_server_stop()

	async def on_server_stop(self):
		return okResponse()

	async def _handle_msg(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
		'''Handle incoming messages.'''
		msg = Cmd(await reader.read())
		addr = writer.get_extra_info('peername')
		print_addr = addr if addr else "localhost"
		self.server_logger.debug(f"Received from {print_addr}")

		if msg.cmd in self.cmd_to_callback:
			self.server_logger.debug(f"Got cmd: {msg.cmd}")
			response = await self.cmd_to_callback[msg.cmd](msg)
		else:
			self.server_logger.warning(f"Got unrecognized command: {msg.cmd.value}")
			response = badResponse()
			response.error = ERRORS.UNRECOGNIZED_COMMAND

		writer.write(response.get_bytes())
		writer.write_eof()
		await writer.drain()

		self.server_logger.debug(f"Closed connection from {print_addr}")
		writer.close()

	async def make_request(self, socket_path: str, cmd: Cmd, expect_response: bool = True) -> Response:
		if os.path.exists(socket_path):
			try:
				reader, writer = await asyncio.open_unix_connection(socket_path)
				writer.write(cmd.get_bytes())
				writer.write_eof()

				if expect_response:
					r = Response(await reader.read())

					writer.close()
					return r
				return okResponse()
			except ConnectionRefusedError:
				self.server_logger.exception(f"Couldn't connect to socket {socket_path}")
				r = badResponse()
				r.error = ERRORS.SOCKET_COULDNT_CONNECT
				return r
		r = badResponse()
		r.error = ERRORS.SOCKET_DOESNT_EXIST
		return r



from pprint import pprint

import pymongo
from kekmonitors.config import Config

if __name__ == "__main__":
	config = Config()

	db = pymongo.MongoClient(config.db_path)
	if config.db_name in db.list_database_names():
		print("Saved items:")
		for item in db["kekmonitors"]["items"].find({}, {"_id": 0}):
			pprint(item)

		print("Registered monitors:")
		for m in db["kekmonitors"]["register"]["monitors"].find({}, {"_id": 0}):
			pprint(m)

		print("Registered scrapers:")
		for m in db["kekmonitors"]["register"]["scrapers"].find({}, {"_id": 0}):
			pprint(m)
	else:
		print("Database does not exist")
